from __future__ import annotations

import ast
from typing import Iterable

from .base import Issue, Rule


class JG05(Issue):
    code = "JG05"
    description = (
        "Condition is too complex which makes it impossible to understand. "
        "Complexity {found} > {allowed}."
    )


class ComplicatedConditionRule(Rule[ast.If]):
    """Rule limitting complexity of IF conditions.

    Too complex conditions are harder to read and can introduce bugs.
    """

    #: Weight of complexity for each op. Default is 1.
    CMP_WEIGHTS: dict[type[ast.cmpop], int] = {
        ast.Eq: 0,
        ast.In: 2,
        ast.NotIn: 2,
    }

    def _calculate_compare_complexity(self, cmp: ast.Compare) -> int:
        return sum(self.CMP_WEIGHTS.get(type(op), 1) for op in cmp.ops)

    def _calculate_complexity(self, test: ast.expr) -> int:
        complexity = 1
        if isinstance(test, ast.UnaryOp):
            complexity += self._calculate_complexity(test.operand)
        elif isinstance(test, ast.BoolOp):
            for value in test.values:
                complexity += self._calculate_complexity(value)
        elif isinstance(test, ast.NamedExpr):
            complexity += self._calculate_complexity(test.value)
        elif isinstance(test, ast.Compare):
            complexity += self._calculate_compare_complexity(test)
        return complexity

    def run(self, node: ast.If) -> Iterable[Issue]:
        issues = []
        complexity = self._calculate_complexity(node.test)

        if complexity > self.options.max_condition_complexity:
            issues.append(
                JG05(
                    lineno=node.lineno,
                    col=node.col_offset,
                    parameters={
                        "found": complexity,
                        "allowed": self.options.max_condition_complexity,
                    },
                )
            )

        return issues



from __future__ import annotations

import argparse
import ast
import dataclasses
from typing import Iterator

from flake8.options.manager import OptionManager

from flake8_jungle.rules import (
    ComplicatedConditionRule,
    FunctionTooLongRule,
    ImportInsideFunctionRule,
    LoggingFormatRule,
    ModelContentOrderRule,
    ModelFieldRule,
    ModelFormRule,
    ModelTooLongRule,
    Rule,
    RuleOptions,
    StructlogFormatRule,
    TooMuchPatchingRule,
    TryExceptPassRule,
)


class JungleStyleVisitor(ast.NodeVisitor):
    """
    Visit the node, and return issues.
    """

    checkers: dict[str, list[type[Rule]]] = {
        "Call": [
            ModelFieldRule,
            LoggingFormatRule,
            StructlogFormatRule,
        ],
        "ClassDef": [
            ModelFormRule,
            ModelContentOrderRule,
            ModelTooLongRule,
        ],
        "FunctionDef": [
            FunctionTooLongRule,
            ImportInsideFunctionRule,
            TooMuchPatchingRule,
        ],
        "AsyncFunctionDef": [
            FunctionTooLongRule,
            ImportInsideFunctionRule,
            TooMuchPatchingRule,
        ],
        "Try": [
            TryExceptPassRule,
        ],
        "If": [
            ComplicatedConditionRule,
        ],
    }

    def __init__(self, options: RuleOptions, *args, **kwargs) -> None:
        super(JungleStyleVisitor, self).__init__(*args, **kwargs)

        self.issues: list = []
        self.options: RuleOptions = options

    def capture_issues_visitor(self, visitor: str, node: ast.AST) -> None:
        for checker in self.checkers[visitor]:
            issues = checker(self.options).run(node)
            if issues:
                self.issues.extend(issues)
        self.generic_visit(node)

    def visit_Call(self, node: ast.Call) -> None:
        self.capture_issues_visitor("Call", node)

    def visit_ClassDef(self, node: ast.ClassDef) -> None:
        self.capture_issues_visitor("ClassDef", node)

    def visit_FunctionDef(self, node: ast.FunctionDef) -> None:
        self.capture_issues_visitor("FunctionDef", node)

    def visit_AsyncFunctionDef(self, node: ast.AsyncFunctionDef) -> None:
        self.capture_issues_visitor("AsyncFunctionDef", node)

    def visit_Try(self, node: ast.Try) -> None:
        self.capture_issues_visitor("Try", node)

    def visit_If(self, node: ast.If) -> None:
        self.capture_issues_visitor("If", node)


class JungleStyleChecker:
    """
    Check common Jungle Style errors
    """

    name = "flake8-jungle"
    version = "1.0.7"

    _options = None
    _option_fields = dataclasses.fields(RuleOptions)
    _default_ignores = ["JG01", "JG03", "JG09"]

    def __init__(self, tree: ast.AST, filename: str) -> None:
        self.tree = tree
        self.filename = filename
        self._rule_options: RuleOptions | None = None

    @property
    def options(self) -> RuleOptions:
        if self._rule_options is None:
            options_kwargs = {}
            for field in self._option_fields:
                value = getattr(self._options, field.name, None)
                if value is not None:
                    options_kwargs[field.name] = value
            self._rule_options = RuleOptions(**options_kwargs)
        return self._rule_options

    @options.setter
    def options(self, value: RuleOptions) -> None:
        self._rule_options = value

    def run(self) -> Iterator[tuple[int, int, str, type]]:
        parser = JungleStyleVisitor(RuleOptions(**vars(self.options)))
        parser.visit(self.tree)

        for issue in parser.issues:
            yield issue.lineno, issue.col, issue.message, JungleStyleChecker

    @classmethod
    def add_options(cls, parser: OptionManager) -> None:
        parser.extend_default_ignore(cls._default_ignores)
        for field in cls._option_fields:
            parser.add_option(
                f'--{field.name.replace("_", "-")}',
                type=field.type,
                default=field.default,
                parse_from_config=True,
            )

    @classmethod
    def parse_options(cls, options: argparse.Namespace) -> None:
        cls._options = options



from __future__ import annotations

import ast
from typing import Iterable

from .base import BaseModelRule, Issue


class JG02(Issue):
    code = "JG02"
    description = (
        "Do not use 'exclude' attribute in ModelForm, list all items "
        "explicitly in 'fields' attribute instead."
    )


class ModelFormRule(BaseModelRule[ast.ClassDef]):
    """Rule forbidding usage of 'exclude' in model forms.

    Using exclude could result in new fields being automatically public
    if new field is added to the model.
    """

    model_name_lookup = "ModelForm"

    def checker_applies(self, node: ast.ClassDef) -> bool:
        for base in node.bases:
            is_model_form = self.is_model_name_lookup(
                base
            ) or self.is_models_name_lookup_attribute(base)
            if is_model_form:
                return True
        return False

    def run(self, node: ast.ClassDef) -> Iterable[Issue]:
        """
        Captures the use of exclude in ModelForm Meta
        """
        issues: list = []

        if not self.checker_applies(node):
            return issues

        for body in node.body:
            if not isinstance(body, ast.ClassDef):
                continue
            for element in body.body:
                if not isinstance(element, ast.Assign):
                    continue
                for target in element.targets:
                    if target.id == "exclude":  # type: ignore
                        issues.append(
                            JG02(
                                lineno=node.lineno,
                                col=node.col_offset,
                            )
                        )
        return issues



from __future__ import annotations

import ast
from typing import Iterable, Union

from .base import Issue, Rule


class JG08(Issue):
    code = "JG08"
    description = (
        "Function or method contains local imports, which should be mostly avoided. "
        "If you are trying to fix curcular dependency issues, the design probably "
        "has some flaws, you should consider refactoring instead."
    )


FunctionType = Union[ast.FunctionDef, ast.AsyncFunctionDef]


class ImportInsideFunctionRule(Rule[FunctionType]):
    """Rule forbidding imports inside functions.

    This can signal bad design and should not be overused.
    """

    def run(self, node: FunctionType) -> Iterable[Issue]:
        issues = []
        contains_import = False

        for line in node.body:
            if isinstance(line, (ast.Import, ast.ImportFrom)):
                contains_import = True
                break

        if contains_import:
            issues.append(
                JG08(
                    lineno=node.lineno,
                    col=node.col_offset,
                )
            )

        return issues



from __future__ import annotations

import ast
from typing import Iterable

from .base import BaseModelRule, Issue


class JG07(Issue):
    code = "JG07"
    description = (
        "Model is too long, split it into services, selectors, or utilities. "
        "Length {found} > {allowed}."
    )


class ModelTooLongRule(BaseModelRule[ast.ClassDef]):
    """Rule limiting length (measured by lines of code) of models.

    Too long models are hard to read, hard to maintain and usually are far from
    following the Domain-Driven Design.
    """

    model_name_lookup = "Model"

    def checker_applies(self, node: ast.ClassDef) -> bool:
        for base in node.bases:
            if self.is_model_name_lookup(base) or self.is_models_name_lookup_attribute(
                base
            ):
                return True
        return False

    def run(self, node: ast.ClassDef) -> Iterable[Issue]:
        issues: list = []

        if not self.checker_applies(node) or node.end_lineno is None:
            return issues

        length = node.end_lineno - node.lineno

        if length > self.options.max_model_length:
            issues.append(
                JG07(
                    lineno=node.lineno,
                    col=node.col_offset,
                    parameters={
                        "found": length,
                        "allowed": self.options.max_model_length,
                    },
                )
            )

        return issues



from __future__ import annotations

import ast
from typing import Iterable

from .base import Issue, Rule

NOT_NULL_TRUE_FIELDS = [
    "CharField",
    "TextField",
    "SlugField",
    "EmailField",
    "FilePathField",
    "URLField",
]


class JG03(Issue):
    code = "JG03"
    description = "Avoid using null=True on string-based fields such as {field}."


class ModelFieldRule(Rule[ast.Call]):
    """Rule forbidding null=True on string-based fields.

    Read Django documentation:
    https://docs.djangoproject.com/en/4.0/ref/models/fields/#null
    """

    def run(self, node: ast.Call) -> Iterable[Issue]:
        issues: list = []

        call_name = self.get_call_name(node)
        if call_name not in NOT_NULL_TRUE_FIELDS:
            return issues

        found_null_true = False
        found_unique_true = False
        found_blank_true = False
        for keyword in node.keywords:
            if keyword.arg == "null" and getattr(keyword.value, "value", False) is True:
                found_null_true = True

            if (
                keyword.arg == "unique"
                and getattr(keyword.value, "value", False) is True
            ):
                found_unique_true = True

            if (
                keyword.arg == "blank"
                and getattr(keyword.value, "value", False) is True
            ):
                found_blank_true = True

            # consider exception for the rule when unique=True and blank=True
            if found_blank_true and found_unique_true:
                return issues

        if found_null_true:
            issues.append(
                JG03(
                    lineno=node.lineno,
                    col=node.col_offset,
                    parameters={"field": call_name},
                )
            )
        return issues



from .base import BaseModelRule, Issue, Rule, RuleOptions
from .jg01_model_content_order import ModelContentOrderRule
from .jg02_model_form import ModelFormRule
from .jg03_model_fields import ModelFieldRule
from .jg04_try_except_pass import TryExceptPassRule
from .jg05_complicated_condition import ComplicatedConditionRule
from .jg06_function_too_long import FunctionTooLongRule
from .jg07_model_too_long import ModelTooLongRule
from .jg08_import_inside_function import ImportInsideFunctionRule
from .jg09_logging_format import LoggingFormatRule
from .jg10_too_much_patching import TooMuchPatchingRule
from .jg11_structlog_format import StructlogFormatRule

__all__ = [
    "Rule",
    "RuleOptions",
    "Issue",
    "BaseModelRule",
    "ModelFieldRule",
    "ModelFormRule",
    "ModelContentOrderRule",
    "TryExceptPassRule",
    "ComplicatedConditionRule",
    "FunctionTooLongRule",
    "ModelTooLongRule",
    "ImportInsideFunctionRule",
    "LoggingFormatRule",
    "StructlogFormatRule",
    "TooMuchPatchingRule",
]



from __future__ import annotations

import ast
from ast import Assign, ClassDef, FunctionDef
from functools import partial
from typing import Callable, Iterable

from .base import BaseModelRule, Issue


class JG01(Issue):
    code = "JG01"
    description = (
        "The order of the model's inner classes, methods, and fields does not follow "
        "the Django Style Guide: {elem_type} should come before {before}."
    )

    def __init__(self, elem, elem_type, before):
        super().__init__(elem.lineno, elem.col_offset, None)
        self.description = self.description.format(
            elem_type=elem_type,
            before=before,
        )


def is_field_declaration(node):
    """
    Verifies that the code is of the form: `field = models.CharField()`,
    matching by the `models` string.
    """
    try:
        return node.value.func.value.id == "models"
    except AttributeError:
        return False


def is_manager_declaration(node: ast.AST) -> bool:
    return (
        isinstance(node, Assign) and getattr(node.targets[0], "id", None) == "objects"
    )


def is_meta_declaration(node):
    return isinstance(node, ClassDef) and node.name == "Meta"


def is_method(node, method_name=None):
    if method_name is None:
        return isinstance(node, FunctionDef)
    return isinstance(node, FunctionDef) and node.name == method_name


class ModelContentOrderRule(BaseModelRule[ast.ClassDef]):
    """Rule which guards the order of model's inner fields, classes, methods
    and so on.
    """

    model_name_lookup = "Model"

    FIELD_DECLARATION = "field declaration"
    MANAGER_DECLARATION = "manager declaration"
    META_CLASS = "Meta class"
    STR_METHOD = "__str__ method"
    SAVE_METHOD = "save method"
    GET_ABSOLUTE_URL_METHOD = "get_absolute_url method"
    CUSTOM_METHOD = "custom method"

    MODEL_CONTENT_TYPE_EXPECTED_ORDER: dict[str, int] = {
        FIELD_DECLARATION: 0,
        MANAGER_DECLARATION: 1,
        META_CLASS: 2,
        STR_METHOD: 3,
        SAVE_METHOD: 4,
        GET_ABSOLUTE_URL_METHOD: 5,
        CUSTOM_METHOD: 6,
    }
    CONTENT_TYPE_CHECKS: list[tuple[Callable, str]] = [
        (is_field_declaration, FIELD_DECLARATION),
        (is_manager_declaration, MANAGER_DECLARATION),
        (is_meta_declaration, META_CLASS),
        (partial(is_method, method_name="__str__"), STR_METHOD),
        (partial(is_method, method_name="save"), SAVE_METHOD),
        (partial(is_method, method_name="get_absolute_url"), GET_ABSOLUTE_URL_METHOD),
        (is_method, CUSTOM_METHOD),
    ]

    def checker_applies(self, node: ast.ClassDef) -> bool:
        for base in node.bases:
            if self.is_model_name_lookup(base) or self.is_models_name_lookup_attribute(
                base
            ):
                return True
        return False

    def run(self, node: ast.ClassDef) -> Iterable[Issue]:
        if not self.checker_applies(node):
            return []

        return self.get_issues(node)

    def get_issues(self, node: ast.ClassDef) -> Iterable[Issue]:
        elements_type_found: list = []
        for element in node.body:
            element_type = self.get_element_type(element)
            if not element_type:
                continue

            element_type_in_wrong_order = self.find_element_type_in_wrong_order(
                element_type, elements_type_found
            )
            if element_type_in_wrong_order:
                yield JG01(
                    element,
                    element_type,
                    element_type_in_wrong_order,
                )
            else:
                elements_type_found.append(element_type)

    def get_element_type(self, element: ast.AST) -> str | None:
        for check, element_type in self.CONTENT_TYPE_CHECKS:
            if check(element):
                return element_type
        return None

    def find_element_type_in_wrong_order(
        self, element_type: str, elements_type_found: list[str]
    ) -> str | None:
        current_element_type_order = self.get_expected_order(element_type)
        for element_type in elements_type_found:
            if self.get_expected_order(element_type) > current_element_type_order:
                return element_type
        return None

    def get_expected_order(self, element_type: str) -> int:
        return self.MODEL_CONTENT_TYPE_EXPECTED_ORDER.get(element_type, 0)



from __future__ import annotations

import ast
from typing import Iterable, Union

from .base import Issue, Rule


class JG06(Issue):
    code = "JG06"
    description = "Function is too long. Length {found} > {allowed}."


FunctionType = Union[ast.FunctionDef, ast.AsyncFunctionDef]


class FunctionTooLongRule(Rule[FunctionType]):
    """Rule limitting length (measured by lines of code) of functions.

    Long functions are harder to read, test and maintain.
    """

    def run(self, node: FunctionType) -> Iterable[Issue]:
        issues: list = []

        if node.end_lineno is None:
            return issues
        length = node.end_lineno - node.lineno

        docstring: str | None = ast.get_docstring(node, clean=False)
        if docstring:
            length -= len(docstring.splitlines())

        if length > self.options.max_function_length:
            issues.append(
                JG06(
                    lineno=node.lineno,
                    col=node.col_offset,
                    parameters={
                        "found": length,
                        "allowed": self.options.max_function_length,
                    },
                )
            )

        return issues



from __future__ import annotations

import ast
import re
from typing import Iterable, Pattern

from .base import Issue, Rule


class JG11(Issue):
    code = "JG11"
    description = (
        "Please use structlog and follow the correct logging style: "
        'logger.{level}("snake_case_message.with_dots", key="value").'
    )


class StructlogFormatRule(Rule[ast.Call]):
    """Rule which enforces correct sturctlog conventions in logger calls.

    See https://www.structlog.org/

    The rule allows the following forms:

        logger.info("payment_task.request_failed", request_id=request.id)
    """

    LOGGER_CALLS: set[str] = {"log", "logger", "logging"}
    MSG_RE: Pattern[str] = re.compile(r"^[a-z\d\._]+$")
    FORMAT_PLACEHOLDER_RE: Pattern[str] = re.compile(r"\{([^\}]*)\}")

    def is_logging_call(self, node: ast.Call) -> bool:
        return (
            isinstance(node.func, ast.Attribute)
            and isinstance(node.func.value, ast.Name)
            and node.func.value.id in self.LOGGER_CALLS
        )

    def is_incorrect_format(self, node: ast.Call) -> bool:
        if not node.args:
            return False
        if len(node.args) > 1:
            return True

        arg0 = node.args[0]

        # simply log.info("msg")
        if isinstance(arg0, ast.Constant):
            msg = arg0.value

        # usage of .format() log.info("{}.msg".format("bar"))
        elif (
            isinstance(arg0, ast.Call)
            and isinstance(arg0.func, ast.Attribute)
            and isinstance(arg0.func.value, ast.Constant)
        ):
            msg = self.FORMAT_PLACEHOLDER_RE.sub("", arg0.func.value.value)

        # usage of f-string log.info(f"{var}.msg")
        elif isinstance(arg0, ast.JoinedStr):
            msg = "".join(
                part.value for part in arg0.values if isinstance(part, ast.Constant)
            )

        # everything else is not checked
        else:
            return False

        if not self.MSG_RE.match(msg):
            return True

        return False

    def run(self, node: ast.Call) -> Iterable[Issue]:
        issues = []

        if self.is_logging_call(node) and self.is_incorrect_format(node):
            issues.append(
                JG11(
                    lineno=node.lineno,
                    col=node.col_offset,
                    parameters={"level": node.func.attr},  # type: ignore
                )
            )

        return issues



from __future__ import annotations

import ast
from dataclasses import dataclass
from typing import Any, Generic, Iterable, TypeVar


class Issue:
    """
    Abstract class for issues.
    """

    code = ""
    description = ""

    def __init__(
        self, lineno: int, col: int, parameters: dict[str, Any] = None
    ) -> None:
        self.parameters = {} if parameters is None else parameters
        self.col = col
        self.lineno = lineno

    @property
    def message(self) -> str:
        """
        Return issue message.
        """
        message = self.description.format(**self.parameters)
        return f"{self.code} {message}"


@dataclass
class RuleOptions:
    """Configuration options for flake8 rules."""

    max_model_length: int = 500
    max_function_length: int = 50
    max_condition_complexity: int = 8
    max_patches_in_test: int = 3


ASTType = TypeVar("ASTType", bound=ast.AST)


class Rule(Generic[ASTType]):
    """
    Abstract class for flake8 rules.
    """

    def __init__(self, options: RuleOptions) -> None:
        self.options = options

    @staticmethod
    def get_call_name(node: ast.Call) -> str | None:
        """
        Return call name for the given node.
        """
        if isinstance(node.func, ast.Attribute):
            return node.func.attr
        elif isinstance(node.func, ast.Name):
            return node.func.id
        return None

    def run(self, node: ASTType) -> Iterable[Issue]:
        """
        Method that runs the checks and returns the issues.
        """
        return NotImplementedError  # type: ignore


class BaseModelRule(Rule[ASTType]):
    """
    Base class for checkers that must lookup for Model like nodes.
    """

    model_name_lookup = ""

    @staticmethod
    def _is_abstract_and_set_to_true(element: ast.stmt) -> bool:
        return (
            isinstance(element, ast.Assign)
            and any(
                target.id == "abstract"
                for target in element.targets
                if isinstance(target, ast.Name)
            )
            and isinstance(element.value, ast.NameConstant)
            and element.value.value is True
        )

    def is_abstract_model(self, base: ast.expr) -> bool:
        """
        Return True if AST node has a Meta class with abstract = True.
        """
        if not hasattr(base, "body"):
            return False

        # look for "class Meta"
        for element in base.body:  # type: ignore
            if isinstance(element, ast.ClassDef) and element.name == "Meta":
                # look for "abstract = True"
                for inner_element in element.body:
                    if self._is_abstract_and_set_to_true(inner_element):
                        return True
        return False

    def is_model_name_lookup(self, base: ast.expr) -> bool:
        """
        Return True if class is defined as the respective model name lookup declaration
        """
        return isinstance(base, ast.Name) and base.id == self.model_name_lookup

    def is_models_name_lookup_attribute(self, base: ast.expr) -> bool:
        """
        Return True if class is defined as the respective model name lookup declaration
        """
        return (
            isinstance(base, ast.Attribute)
            and isinstance(base.value, ast.Name)
            and base.value.id == "models"
            and base.attr == self.model_name_lookup
        )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['flake8_jungle', 'flake8_jungle.rules']

package_data = \
{'': ['*']}

install_requires = \
['flake8>=3.8.4']

entry_points = \
{'flake8.extension': ['JG = flake8_jungle:JungleStyleChecker']}

setup_kwargs = {
    'name': 'flake8-jungle',
    'version': '1.0.7',
    'description': 'Plugin to lint various issues in code.',
    'long_description': '# flake8-jungle\n\nA flake8 plugin to detect bad practices in projects. This plugin is based on [flake8-django](https://github.com/rocioar/flake8-django/).\n\n## Installation\n\nInstall from pip with:\n\n```\n$ pip install flake8-jungle\n```\n\n## `pre-commit` example\n\n```yaml\n  - repo: https://github.com/pycqa/flake8\n    rev: 4.0.1\n    hooks:\n      - id: flake8\n        additional_dependencies: [\'flake8-jungle==VERSION\']\n        args: [\'--max-condition-complexity=8\']\n```\n\n## List of Rules\n\n| Rule | Description | Configuration |\n| ---- | ----------- | ------------- |\n| `JG02` | Do not use `exclude` attribute in `ModelForm`, list all items explicitly in `fields` attribute instead. | |\n| `JG04` | Exceptions should never pass silently, add logging or comment at least. | |\n| `JG05` | Condition is too complex which makes it hard to understand. | `--max-condition-complexity` |\n| `JG06` | Function is too long. | `--max-function-length` |\n| `JG07` | Model is too long, split it into services, selectors, or utilities. | `--max-model-length` |\n| `JG08` | Function or method contains local imports, which should be mostly avoided. If you are trying to fix curcular dependency issues, the design probably has some flaws, you should consider refactoring instead. |\n| `JG10` | Too much patching in tests. Consider changing your design to utilize Dependency Injection and fakes. | `--max-patches-in-test` |\n| `JG11` | Please use structlog and follow the correct logging style: `logger.info("snake_case_message.with_dots", key="value")`. | |\n\nThe following rules are disabled by default:\n\n| Rule | Description | Configuration |\n| ---- | ----------- | ------------- |\n| `JG01` | The order of the model\'s inner classes, methods, and fields does not follow the [Django Style Guide](https://github.com/HackSoftware/Django-Styleguide). | |\n| `JG03` | Avoid using `null=True` on string-based fields such as `CharField` and `TextField`. | |\n| `JG09` | Incorrect logging format, please use the following syntax: `logger.info("MESSAGE %(arg1)s", {"arg1": "value1"})`. | |\n\nTo enable optional rules you can use the `--select` parameter. It\'s default values are: `E,F,W,C90`.\n\nFor example, if you wanted to enable `JG10`, you could call `flake8` in the following way:\n\n```bash\nflake8 --select=E,F,W,C90,JG,JG10\n```\n\n## Testing\n\nflake8-jungle uses pytest for tests. To run them use:\n\n```\n$ poetry install\n$ poetry run pytest tests\n```\n',
    'author': 'Twisto Platform Team',
    'author_email': 'platform@twisto.cz',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/TwistoPayments/flake8-jungle',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from __future__ import annotations

import ast
from typing import Iterable, Union

from .base import Issue, Rule


class JG10(Issue):
    code = "JG10"
    description = (
        "Too much patching in tests ({found} > {allowed}). "
        "Consider changing your design to utilize Dependency Injection and fakes."
    )


FunctionType = Union[ast.FunctionDef, ast.AsyncFunctionDef]


class TooMuchPatchingRule(Rule[FunctionType]):
    """Rule which counts number of `unittest.patch(...)` calls in tests.

    The rule suggests different approach if patching in tests is being overused.
    There are several reasons why patching might not be the best option:

    - it may signal that the code is violating SOLID principles
    - patching makes tests too dependent on module paths, function and class names
    - patching makes tests harder to maintain
    """

    PATCH_NAMES: set[str] = {"patch"}

    def is_patch_call(self, call: ast.AST) -> bool:
        if not isinstance(call, ast.Call):
            return False

        name = None
        if isinstance(call.func, ast.Attribute):
            name = call.func.attr
        elif isinstance(call.func, ast.Name):
            name = call.func.id
        return name in self.PATCH_NAMES

    def is_test_function(self, node: FunctionType) -> bool:
        return node.name.startswith("test_")

    def is_patch_expr(self, node: ast.AST) -> bool:
        return isinstance(node, ast.Expr) and self.is_patch_call(node.value)

    def is_patch_with(self, node: ast.AST) -> bool:
        return isinstance(node, ast.With)

    def count_patch_with_items(self, items: list[ast.withitem]) -> int:
        number_of_patch_items = 0
        for item in items:
            if self.is_patch_call(item.context_expr):
                number_of_patch_items += 1
        return number_of_patch_items

    def count_number_of_patches(self, node: FunctionType) -> int:
        number_of_patch_calls = 0

        for line in node.body:
            if self.is_patch_expr(line):
                number_of_patch_calls += 1
            elif self.is_patch_with(line):
                number_of_patch_calls += self.count_patch_with_items(
                    line.items  # type: ignore
                )

        for item in node.decorator_list:
            if self.is_patch_call(item):
                number_of_patch_calls += 1

        return number_of_patch_calls

    def run(self, node: FunctionType) -> Iterable[Issue]:
        issues: list = []

        if not self.is_test_function(node):
            return issues

        number_of_patches = self.count_number_of_patches(node)
        if number_of_patches > self.options.max_patches_in_test:
            issues.append(
                JG10(
                    lineno=node.lineno,
                    col=node.col_offset,
                    parameters={
                        "found": number_of_patches,
                        "allowed": self.options.max_patches_in_test,
                    },
                )
            )

        return issues



from __future__ import annotations

import ast
from typing import Iterable

from .base import Issue, Rule


class JG09(Issue):
    code = "JG09"
    description = (
        "Incorrect logging format, please use the following syntax: "
        'logger.{level}("MESSAGE %(arg1)s", {{"arg1": "value1"}}).'
    )


class LoggingFormatRule(Rule[ast.Call]):
    """Rule which forbids usage of `.format()`, `f-string` and `% (var,)` usage
    in logging calls.

    The only allowed forms are:

        logger.info("Request %s did not succed.", request.correlation_id)
        logger.info("Request %(id)s did not succed.", {"id": request.correlation_id})
    """

    LOGGER_CALLS: set[str] = {"log", "logger", "logging"}

    def is_logging_call(self, node: ast.Call) -> bool:
        return (
            isinstance(node.func, ast.Attribute)
            and isinstance(node.func.value, ast.Name)
            and node.func.value.id in self.LOGGER_CALLS
        )

    def is_incorrect_format(self, node: ast.Call) -> bool:
        if len(node.args) == 1:
            not_allowed = (
                ast.Call,  # represents "{}".format("string")
                ast.BinOp,  # represents "%s" % ("string",)
                ast.JoinedStr,  # represents f"{variable}"
            )
            if isinstance(node.args[0], not_allowed):
                return True
        return False

    def run(self, node: ast.Call) -> Iterable[Issue]:
        issues = []

        if self.is_logging_call(node) and self.is_incorrect_format(node):
            issues.append(
                JG09(
                    lineno=node.lineno,
                    col=node.col_offset,
                    parameters={"level": node.func.attr},  # type: ignore
                )
            )

        return issues



from __future__ import annotations

import ast
from typing import Iterable

from .base import Issue, Rule


class JG04(Issue):
    code = "JG04"
    description = (
        "Exceptions should never pass silently, add logging or comment at least."
    )


class TryExceptPassRule(Rule[ast.Try]):
    """Rule forbidding silent exception handling.

    Even though silent handling is Pythonic in some cases, it can often result
    in hiding important information.
    """

    def _is_pass(self, node: ast.stmt) -> bool:
        return isinstance(node, ast.Pass)

    def _is_return_none(self, node: ast.stmt) -> bool:
        if isinstance(node, ast.Return):
            if node.value is None:
                return True
            if isinstance(node.value, ast.Constant) and node.value.value is None:
                return True
        return False

    def run(self, node: ast.Try) -> Iterable[Issue]:
        issues = []

        for handler in node.handlers:
            if len(nodes := handler.body) != 1:
                continue
            if self._is_return_none(nodes[0]) or self._is_pass(nodes[0]):
                issues.append(
                    JG04(
                        lineno=nodes[0].lineno,
                        col=nodes[0].col_offset,
                    )
                )

        return issues
